;;;
;;; An implementation of the Memory Array Redcode Simulator (MARS)
;;;

(import (chicken io)
        (chicken string)
        (chicken random)
        (chicken condition)
        (chicken process-context)
        matchable)


;;; Instructions
;;

(define (make-instr opcode modifier A-mode A-num B-mode B-num)
  (lambda args
    (match args
      (('opcode) opcode)
      (('modifier) modifier)
      (('A-mode) A-mode)
      (('A-num) A-num)
      (('B-mode) B-mode)
      (('B-num) B-num)
      (('make-copy) (make-instr opcode modifier A-mode A-num B-mode B-num))
      (('set! 'opcode x) (set! opcode x))
      (('set! 'modifier x) (set! modifier x))
      (('set! 'A-mode x) (set! A-mode x))
      (('set! 'A-num x) (set! A-num x))
      (('set! 'B-mode x) (set! B-mode x))
      (('set! 'B-num x) (set! B-num x))
      (('set-from! other)
       (set! opcode (other 'opcode))
       (set! modifier (other 'modifier))
       (set! A-mode (other 'A-mode))
       (set! A-num (other 'A-num))
       (set! B-mode (other 'B-mode))
       (set! B-num (other 'B-num)))
      (('->string)
       (conc opcode
             "." modifier
             " " (mode->string A-mode) A-num
             " " (mode->string B-mode) B-num))
      (else
       (error "Invalid instr arguments" args)))))

(define (mode->string mode)
  (case mode
    ((immediate) "#")
    ((direct) "$")
    ((indirect-A) "*")
    ((indirect-B) "@")
    ((pre-indirect-A) "{")
    ((pre-indirect-B) "<")
    ((post-indirect-A) "}")
    ((post-indirect-B) ">")
    (else
     (error "Unknown mode."))))

(define initial-instruction
  (make-instr 'DAT 'F 'immediate 0 'immediate 0))


;;; Memory setup and addressing
;;

(define (make-core core-size initial-instr)
  (let ((core-vec (make-vector core-size '()))
        (names-vec (make-vector core-size '())))
    (let loop ((i 0))
      (unless (>= i core-size)
        (vector-set! core-vec i (initial-instr 'make-copy))
        (loop (+ i 1))))
    (lambda args
      (match args
        ((i 'set-from! j n)
         ((vector-ref core-vec i) 'set-from! (vector-ref core-vec j))
         (vector-set! names-vec i n))
        ((i 'set-from-instr! instr n)
         ((vector-ref core-vec i) 'set-from! instr)
         (vector-set! names-vec i n))
        ((i 'set! v x n)
         ((vector-ref core-vec i) 'set! v x)
         (vector-set! names-vec i n))
        ((i 'name) (vector-ref names-vec i))
        ((i v) ((vector-ref core-vec i) v))
        (('dump)
         (let loop ((i 0))
           (unless (>= i core-size)
             (print* i ":\t" ((vector-ref core-vec i) '->string))
             (let ((n (vector-ref names-vec i)))
               (unless (null? n)
                 (print* "\t;" n)))
             (print)
             (loop (+ i 1)))))))))

(define (addr+ . args)
  (foldl (lambda (a b)
           (modulo (+ a b core-size) core-size))
         0 args))

;;; Programmes and task queues
;;

(define (make-prog name instrs offset)
  (list name instrs offset))

(define (prog-name prog) (list-ref prog 0))
(define (prog-instrs prog) (list-ref prog 1))
(define (prog-offset prog) (list-ref prog 2))

(define (install-prog core prog addr)
  (let loop ((ptr addr)
             (instrs (prog-instrs prog)))
    (unless (null? instrs)
      (core ptr 'set-from-instr! (car instrs) (prog-name prog))
      (loop (addr+ ptr 1) (cdr instrs))))
  (make-queue (prog-name prog)
              (addr+ addr (prog-offset prog))))

(define (can-install-prog? core prog-len addr)
  (let loop ((ptr addr)
             (remaining prog-len))
    (if (= remaining 0)
        #t
        (if (null? (core ptr 'name))
            (loop (addr+ ptr 1)
                  (- remaining 1))
            #f))))

(define (install-progs core progs)
  (let loop ((queues '())
             (progs-left progs))
    (if (null? progs-left)
        queues
        (let ((addr (pseudo-random-integer core-size))
              (prog (car progs-left)))
          (if (can-install-prog? core (length (prog-instrs prog)) addr)
              (loop (cons (install-prog core prog addr) queues)
                    (cdr progs-left))
              (loop queues progs-left))))))

(define (make-queue name ptr)
  (list name ptr))

(define (queue-owner queue) (car queue))
(define (queue-ptrs queue) (cdr queue))

(define (queue-set-ptrs! queue ptrs)
  (set-cdr! queue ptrs))


;;; Executive function
;;

(define (run core queues steps-left)
  (cond
   ((<= steps-left 0) queues)      ;Tie between remaining players
   ;; ((<= (length queues) 1) queues) ;There's only one player left who thus wins
   (else
    (let* ((queue (car queues))
           (remaining-queues (cdr queues))
           (ptrs (queue-ptrs queue))
           (new-ptrs (execute-instr core (car ptrs) (queue-owner queue))))
      (if (null? new-ptrs)
          (run remaining-queues (- steps-left 1))
          (begin
            (queue-set-ptrs! queue (append (cdr ptrs) new-ptrs))
            (run core (append remaining-queues (list queue)) (- steps-left 1))))))))

(define (execute-instr core ptr name)
  (let* ((A-ptr (eval-operand core (core ptr 'A-mode) (core ptr 'A-num) ptr name))
         (B-ptr (eval-operand core (core ptr 'B-mode) (core ptr 'B-num) ptr name))
         (modifier (core ptr 'modifier)))
    (case (core ptr 'opcode)
      ((DAT)
       '()) ;Game over, man, game over!
      ((MOV)
       (if (eq? modifier 'I)
           (core B-ptr 'set-from! A-ptr name)
           (combine-and-store core A-ptr B-ptr modifier name (lambda (x y) y)))
       (list (addr+ ptr 1)))
      ((ADD)
       (combine-and-store core A-ptr B-ptr modifier name addr+)
       (list (addr+ ptr 1)))
      ((SUB)
       (combine-and-store core A-ptr B-ptr modifier name
                          (lambda (x y) (addr+ x (- y))))
       (list (addr+ ptr 1)))
      ((MUL)
       (combine-and-store core A-ptr B-ptr modifier name
                          (lambda (x y) (modulo (* (addr+ x core-size)
                                                   (addr+ y core-size))
                                                core-size)))
       (list (addr+ ptr 1)))
      ((DIV)
       (condition-case 
           (begin
             (combine-and-store core A-ptr B-ptr modifier name
                                (lambda (x y) (quotient (addr x core-size)
                                                        (addr y core-size))))
             (list (addr+ ptr 1)))
         ((exn arithmetic) '())))
      ((MOD)
       (condition-case
           (begin
             (combine-and-store core A-ptr B-ptr modifier name
                                (lambda (x y) (remainder (addr x core-size)
                                                         (addr y core-size))))
             (list (addr+ ptr 1)))
         ((exn arithmetic) '())))
      ((JMP)
       (list (addr+ ptr (core A-ptr 'A-num))))
      ((JMZ)
       (list (addr+ ptr (if (instr-zero? B-ptr modifier #f name)
                            (core A-ptr 'A-num)
                            1))))
      ((JMN)
       (list (addr+ ptr (if (not (instr-zero? B-ptr modifier #f name))
                            (core A-ptr 'A-num)
                            1))))
      ((DJN)
       (list (addr+ ptr (if (not (instr-zero? B-ptr modifier #t name))
                            (core A-ptr 'A-num)
                            1))))
      ((SEQ CMP)
       (list (addr+ ptr (if (compare-instrs core A-ptr B-ptr modifier =) 2 1))))
      ((SNE)
       (list (addr+ ptr (if (compare-instrs core A-ptr B-ptr modifier =) 1 2))))
      ((SLT)
       (list (addr+ ptr (if (compare-instrs core A-ptr B-ptr modifier <) 2 1))))
      ((SPL)
       (list (addr+ ptr 1) (addr+ ptr (core A-ptr 'A-num))))
      ((NOP)
       (list (addr+ ptr 1)))
      (else
       (error "Unrecognised opcode" (instr 'opcode))))))

(define (compare-instrs core A-ptr B-ptr modifier test)
  (case modifier
    ((A) (test (core A-ptr 'A-num) (core B-ptr 'A-num)))
    ((B) (test (core A-ptr 'B-num) (core B-ptr 'B-num)))
    ((AB) (test (core A-ptr 'A-num) (core B-ptr 'B-num)))
    ((BA) (test (core A-ptr 'B-num) (core B-ptr 'A-num)))
    ((F) (and
          (test (core A-ptr 'A-num) (core B-ptr 'A-num))
          (test (core A-ptr 'B-num) (core B-ptr 'B-num))))
    ((X) (and
          (test (core A-ptr 'A-num) (core B-ptr 'B-num))
          (test (core A-ptr 'B-num) (core B-ptr 'A-num))))
    ((I) (and
          (if (eq? test =)
              (and
               (eq? (core A-ptr 'opcode) (core B-ptr 'opcode))
               (eq? (core A-ptr 'modifier) (core B-ptr 'modifier))
               (eq? (core A-ptr 'A-mode) (core B-ptr 'B-mode))
               (eq? (core A-ptr 'B-mode) (core B-ptr 'A-mode)))
              #t)
          (test (core A-ptr 'A-num) (core B-ptr 'B-num))
          (test (core A-ptr 'B-num) (core B-ptr 'A-num))))))

(define (instr-zero? core ptr modifier decrement name)
  (case modifier
    ((A AB)
     (if decrement (core ptr 'set! 'A-num (addr+ (core ptr 'A-num) -1) name))
     (= 0 (core ptr 'A-num)))
    ((A AB)
     (if decrement (core ptr 'set! 'B-num (addr+ (core ptr 'B-num) -1) name))
     (= 0 (core ptr 'B-num)))
    ((X I F)
     (if decrement
         (begin
           (core ptr 'set! 'A-num (addr+ (core ptr 'A-num) -1) name)
           (core ptr 'set! 'B-num (addr+ (core ptr 'B-num) -1) name)))
     (and (= 0 (core ptr 'A-num))
          (= 0 (core ptr 'B-num))))))

(define (combine-and-store core A-ptr B-ptr modifier name f)
  (case modifier
    ((A) (core B-ptr 'set! 'A-num
               (f (core B-ptr 'A-num) (core A-ptr 'A-num)) name))
    ((B) (core B-ptr 'set! 'B-num
               (f (core B-ptr 'B-num) (core A-ptr 'B-num)) name))
    ((AB) (core B-ptr 'set! 'B-num
                (f (core B-ptr 'B-num) (core A-ptr 'A-num)) name))
    ((BA) (core B-ptr 'set! 'A-num
                (f (core B-ptr 'A-num) (core A-ptr 'B-num)) name))
    ((F I) (core B-ptr 'set! 'A-num
                 (f (core B-ptr 'A-num) (core A-ptr 'A-num)) name)
           (core B-ptr 'set! 'B-num
                 (f (core B-ptr 'B-num) (core A-ptr 'B-num)) name))
    ((X) (core B-ptr 'set! 'A-num
               (f (core B-ptr 'A-num) (core A-ptr 'B-num)) name)
         (core B-ptr 'set! 'B-num
               (f (core B-ptr 'B-num) (core A-ptr 'A-num)) name))))

(define (eval-operand core mode num ptr name)
  (addr+ ptr
         (case mode
           ((immediate) 0)
           ((direct) num)
           ((indirect-A) (addr+ num (core (addr+ ptr num) 'A-num)))
           ((indirect-B) (addr+ num (core (addr+ ptr num) 'B-num)))
           ((pre-indirect-A)
            (let ((aux-ptr (addr+ ptr num)))
              (core aux-ptr 'set! 'A-num (addr+ -1 (core aux-ptr 'A-num)) name)
              (addr+ num (core aux-ptr 'A-num))))
           ((pre-indirect-B)
            (let ((aux-ptr (addr+ ptr num)))
              (core aux-ptr 'set! 'B-num (addr+ -1 (core aux-ptr 'B-num)) name)
              (addr+ num (core aux-ptr 'B-num))))
           ((post-indirect-A)
            (let* ((aux-ptr (addr+ ptr num))
                   (old-A-num (core aux-ptr 'A-num)))
              (core aux-ptr 'set! 'A-num (addr+ 1 (core aux-ptr 'A-num)) name)
              (addr+ num old-A-num)))
           ((post-indirect-B)
            (let* ((aux-ptr (addr+ ptr num))
                   (old-B-num (core aux-ptr 'B-num)))
              (core aux-ptr 'set! 'B-num (addr+ 1 (core aux-ptr 'B-num)) name)
              (addr+ num old-B-num)))
           (else
            (error "Unrecognized mode" mode)))))

;;; Main procedure
;; 

;;; TEST CODE

(define addressing-test
  (make-prog 'at (list
                   (make-instr 'DAT 'F 'immediate 42 'immediate 53)
                   (make-instr 'DAT 'F 'immediate 123 'immediate 256)
                   (make-instr 'MOV 'A 'indirect-B 4 'direct 7)
                   (make-instr 'NOP 'I 'immediate 0 'immediate 0)
                   (make-instr 'NOP 'I 'immediate 0 'immediate 0)
                   (make-instr 'NOP 'I 'immediate 0 'immediate 0)
                   (make-instr 'DAT 'F 'immediate -5 'immediate -6)) 2))

(define imp
  (make-prog 'imp (list (make-instr 'MOV 'I 'direct 0 'direct 1)) 0))

(define dwarf
  (make-prog 'dwarf (list
                     (make-instr 'DAT 'F 'immediate 0 'immediate -1)
                     (make-instr 'ADD 'AB 'immediate 5 'direct -1)
                     (make-instr 'MOV 'I 'direct -2 'indirect-B -2)
                     (make-instr 'JMP 'I 'immediate -2 'immediate 0)) 1))

(define core (make-core 20 (make-instr 'DAT 'F 'immediate 0 'immediate 0)))
(define queues (install-progs core (list dwarf imp)))
